import logging
import subprocess

from docker.errors import ImageNotFound, APIError  # type: ignore

from biolib import utils
from biolib.typing_utils import TypedDict
from biolib.biolib_docker_client import BiolibDockerClient
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger_no_user_data
from biolib.compute_node.job_worker.cache_state import DockerImageCacheState, DockerCacheStateError
from biolib.compute_node.job_worker.cache_types import DockerImageInfo, DockerImageCacheStateDict, DockerAuthConfig, \
    UuidStr


class DiskUsage(TypedDict):
    image_storage_available_in_bytes: int
    max_image_capacity_in_bytes: int
    used_in_bytes: int


class DockerImageCache:
    def __init__(self):
        if not utils.IS_RUNNING_IN_CLOUD:
            raise BioLibError('Using DockerImageCache outside Cloud is not supported')

        self._docker_client = BiolibDockerClient().get_docker_client()
        self._docker_data_dir = self._docker_client.info()['DockerRootDir']

    def _get_current_disk_usage_stats(self) -> DiskUsage:
        disk_usage_command = ['df', '--block-size=1', self._docker_data_dir]
        logger_no_user_data.debug(f'Running {disk_usage_command}...')
        df_output = subprocess.run(
            disk_usage_command,
            capture_output=True,
            check=True,
            timeout=3,
        ).stdout.decode()
        logger_no_user_data.debug(df_output)
        df_output_lines = [s.split() for s in df_output.splitlines()]

        # Expect the filesystem to be row 1
        disk = df_output_lines[1]
        # Expect column 1 to be total disk size and column 2 to be disk space used
        total_size_in_bytes = int(disk[1])
        used_in_bytes = int(disk[2])

        # Allow 80% to be used for images
        max_image_capacity_in_bytes = total_size_in_bytes * 80 // 100
        available_for_images_in_bytes = max(0, max_image_capacity_in_bytes - used_in_bytes)

        disk_usage = DiskUsage(
            image_storage_available_in_bytes=available_for_images_in_bytes,
            max_image_capacity_in_bytes=max_image_capacity_in_bytes,
            used_in_bytes=used_in_bytes,
        )
        logger_no_user_data.debug(f'Disk usage for Docker: {disk_usage}')
        return disk_usage

    def _clear_space_for_image(self, estimated_image_size_bytes: int, cache_state: DockerImageCacheStateDict):
        for _ in range(100):
            if not self._has_space_to_pull_image(estimated_image_size_bytes, cache_state):
                self._remove_least_recently_used_image(cache_state)
            else:
                return

        raise DockerCacheStateError('Failed to free space for Docker image')

    def get(
            self,
            image_uri: str,
            estimated_image_size_bytes: int,
            pull_auth_config: DockerAuthConfig,
            job_id: str,
    ) -> None:
        try:
            with DockerImageCacheState() as cache_state:
                if image_uri not in cache_state:
                    raise ImageNotFound('Image not found in cache')

                self._docker_client.images.get(image_uri)
                cache_state[image_uri]['last_used_at'] = DockerImageCacheState.get_timestamp_now()
                cache_state[image_uri]['active_jobs'].append(job_id)

        except ImageNotFound:
            disk_stats = self._get_current_disk_usage_stats()
            if estimated_image_size_bytes > disk_stats['max_image_capacity_in_bytes']:
                logger_no_user_data.error(
                    f'Image {image_uri} with size: {estimated_image_size_bytes} is bigger than the max cache size'
                )
                raise DockerCacheStateError(  # pylint: disable=raise-missing-from
                    'Image is bigger than the max cache size'
                )

            with DockerImageCacheState() as cache_state:
                if not self._has_space_to_pull_image(estimated_image_size_bytes, cache_state):
                    self._clear_space_for_image(estimated_image_size_bytes, cache_state)

                cache_state[image_uri] = DockerImageInfo(
                    last_used_at=DockerImageCacheState.get_timestamp_now(),
                    uri=image_uri,
                    state='pulling',
                    estimated_image_size_bytes=estimated_image_size_bytes,
                    active_jobs=[job_id]
                )

            logger_no_user_data.debug(f'Image {image_uri} not found in cache. Pulling...')
            try:
                self._docker_client.images.pull(image_uri, auth_config=pull_auth_config)
            except Exception as error:
                logger_no_user_data.error(f'Could not pull image due to {error}')

                # Remove image from cache_state
                with DockerImageCacheState() as cache_state:
                    if image_uri in cache_state:
                        cache_state.pop(image_uri)

                raise error

            with DockerImageCacheState() as cache_state:
                cache_state[image_uri]['state'] = 'ready'

    @staticmethod
    def detach_job(image_uri: str, job_id: UuidStr) -> None:
        try:
            with DockerImageCacheState() as cache_state:
                if image_uri in cache_state and job_id in cache_state[image_uri]['active_jobs']:
                    cache_state[image_uri]['active_jobs'].remove(job_id)

        except Exception as error:  # pylint: disable=broad-except
            logging.error('Could not remove job from image cache')
            logging.error(error)

    def _remove_least_recently_used_image(self, cache_state: DockerImageCacheStateDict) -> None:
        cached_images = [image for image in cache_state.values() if image['state'] == 'ready']
        images_sorted_by_least_recently_used = sorted(cached_images, key=lambda image: image['last_used_at'])

        for image in images_sorted_by_least_recently_used:
            logger_no_user_data.debug(f"Removing image: {image['uri']}")

            # Only remove images that has no active jobs
            if image['active_jobs']:
                logger_no_user_data.debug(
                    f"Skipping removal of image {image['uri']} as it has the active jobs {image['active_jobs']}"
                )
                continue

            try:
                self._docker_client.api.remove_image(image=image['uri'])
            except APIError as error:
                logger_no_user_data.error(
                    f'Could not remove image due to {error}... Skipping removal of this image.'
                )
                continue  # Image is in use or cannot be removed at this time

            cache_state.pop(image['uri'])
            break

    def _has_space_to_pull_image(self, estimated_image_size_bytes: int, cache_state: DockerImageCacheStateDict) -> bool:
        logger_no_user_data.debug('Calculating cache metrics...')
        size_of_images_being_pulled = sum(
            [image['estimated_image_size_bytes'] for image in cache_state.values() if image['state'] == 'pulling']
        )
        disk_stats = self._get_current_disk_usage_stats()
        cache_space_remaining = disk_stats['image_storage_available_in_bytes'] - size_of_images_being_pulled

        logger_no_user_data.debug(
            f'Needed current image: {estimated_image_size_bytes}. '
            f'Cache remaining: {cache_space_remaining}. '
            f'Total pulling: {size_of_images_being_pulled}. '
        )

        return bool(cache_space_remaining > estimated_image_size_bytes)



import math
import multiprocessing
import multiprocessing.pool
import os
import time
from urllib.parse import urlparse

import requests

from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_api_client.auth import BearerAuth
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger, logger_no_user_data
from biolib.typing_utils import TypedDict, List, Iterator, Tuple, Optional, Dict


def get_chunk_iterator_from_bytes(byte_buffer: bytes, chunk_size_in_bytes: int = 50_000_000) -> Iterator[bytes]:
    chunk_count = math.ceil(len(byte_buffer) / chunk_size_in_bytes)
    for chunk_number in range(chunk_count):
        start = chunk_size_in_bytes * chunk_number
        stop = start + chunk_size_in_bytes
        yield byte_buffer[start:stop]


def get_chunk_iterator_from_file_object(file_object, chunk_size_in_bytes: int = 50_000_000) -> Iterator[bytes]:
    while True:
        data = file_object.read(chunk_size_in_bytes)
        if not data:
            break
        yield data


class RequestOptions(TypedDict):
    headers: Optional[Dict[str, str]]
    requires_biolib_auth: bool
    url: str


class _PartMetadata(TypedDict):
    ETag: str
    PartNumber: int


_UploadChunkInputType = Tuple[int, bytes]
_UploadChunkReturnType = Tuple[_PartMetadata, int]


class MultiPartUploader:

    def __init__(
            self,
            complete_upload_request: RequestOptions,
            get_presigned_upload_url_request: RequestOptions,
            start_multipart_upload_request: Optional[RequestOptions] = None,
    ):
        self._complete_upload_request = complete_upload_request
        self._get_presigned_upload_url_request = get_presigned_upload_url_request
        self._start_multipart_upload_request = start_multipart_upload_request
        self._bytes_uploaded: int = 0

    def upload(self, payload_iterator: Iterator[bytes], payload_size_in_bytes: int) -> None:
        parts: List[_PartMetadata] = []

        iterator_with_index: Iterator[_UploadChunkInputType] = enumerate(payload_iterator, 1)  # type: ignore
        logger_no_user_data.debug(f'Starting multipart upload of payload with size {payload_size_in_bytes} bytes')

        if self._start_multipart_upload_request:
            requires_biolib_auth = self._start_multipart_upload_request['requires_biolib_auth']
            start_multipart_upload = requests.post(
                auth=BearerAuth(BiolibApiClient.get().access_token) if requires_biolib_auth else None,
                headers=self._start_multipart_upload_request['headers'],
                timeout=30,
                url=self._start_multipart_upload_request['url'],
            )
            if start_multipart_upload.ok:
                logger_no_user_data.debug('Multipart upload started')
            else:
                logger_no_user_data.debug(
                    f'Failed to start multipart upload got response status: {start_multipart_upload.status_code}'
                )
                raise Exception('Failed to start multipart upload')

        # if multiprocessing start method is spawn or we are running in a daemon process,
        # multiprocessing.Pool may fail when called from script
        if multiprocessing.get_start_method() == 'spawn' or multiprocessing.current_process().daemon:
            logger_no_user_data.debug('Uploading multipart from main process...')
            for chunk in iterator_with_index:
                upload_chunk_response = self._upload_chunk(chunk)
                self._update_progress_bar_and_parts(
                    upload_chunk_response=upload_chunk_response,
                    parts=parts,
                    payload_size_in_bytes=payload_size_in_bytes
                )
        else:
            # use 16 cores, unless less is available
            process_pool = multiprocessing.pool.ThreadPool(processes=min(16, multiprocessing.cpu_count() - 1))
            try:
                response: _UploadChunkReturnType
                for response in process_pool.imap(self._upload_chunk, iterator_with_index):
                    self._update_progress_bar_and_parts(
                        upload_chunk_response=response,
                        parts=parts,
                        payload_size_in_bytes=payload_size_in_bytes
                    )
            finally:
                logger_no_user_data.debug('Multipart upload closing process pool...')
                process_pool.close()

        requires_biolib_auth = self._complete_upload_request['requires_biolib_auth']
        if requires_biolib_auth:
            BiolibApiClient.refresh_auth_token()

        logger_no_user_data.debug(f'Uploaded {len(parts)} parts, now calling complete upload...')
        for index in range(3):
            try:
                complete_upload_response = requests.post(
                    auth=BearerAuth(BiolibApiClient.get().access_token) if requires_biolib_auth else None,
                    headers=self._complete_upload_request['headers'],
                    json={'parts': parts, 'size_bytes': self._bytes_uploaded},
                    timeout=30,
                    url=self._complete_upload_request['url'],
                )
                if complete_upload_response.ok:
                    logger_no_user_data.debug('Multipart upload completed returning')
                    return

                logger_no_user_data.warning(
                    f'Failed to complete multipart upload got response status {complete_upload_response.status_code}. '
                    f'Retrying...'
                )

            except Exception as error:  # pylint: disable=broad-except
                logger_no_user_data.warning('Encountered error when completing multipart upload. Retrying...')
                logger.debug(f'Multipart complete error: {error}')
                time.sleep(index * index + 2)

        raise BioLibError('Max retries hit, when completing multipart upload')

    def _upload_chunk(self, _input: _UploadChunkInputType) -> _UploadChunkReturnType:
        part_number, chunk = _input
        requires_biolib_auth = self._get_presigned_upload_url_request['requires_biolib_auth']

        for index in range(20):  # will fail after approximately sum_i(i^2+2) = 41 min if range (20)
            if requires_biolib_auth:
                BiolibApiClient.refresh_auth_token()

            logger_no_user_data.debug(f'Uploading part number {part_number} with size {len(chunk)} bytes...')
            try:
                logger_no_user_data.debug(f'Getting upload URL for chunk {part_number}...')
                get_url_response = requests.get(
                    auth=BearerAuth(BiolibApiClient.get().access_token) if requires_biolib_auth else None,
                    headers=self._get_presigned_upload_url_request['headers'],
                    params={'part_number': part_number},
                    timeout=30,
                    url=self._get_presigned_upload_url_request['url'],
                )
                if not get_url_response.ok:
                    raise Exception(
                        f'Failed to get upload URL for part {part_number} got response status code '
                        f'{get_url_response.status_code}'
                    )

                presigned_upload_url = get_url_response.json()['presigned_upload_url']

                app_caller_proxy_job_storage_base_url = os.getenv('BIOLIB_CLOUD_JOB_STORAGE_BASE_URL', '')
                if app_caller_proxy_job_storage_base_url:
                    # Done to hit App Caller Proxy when uploading result from inside an app
                    parsed_url = urlparse(presigned_upload_url)
                    presigned_upload_url = \
                        f'{app_caller_proxy_job_storage_base_url}{parsed_url.path}?{parsed_url.query}'

                put_chunk_response = requests.put(url=presigned_upload_url, data=chunk, timeout=300)

                if put_chunk_response.ok:
                    return _PartMetadata(PartNumber=part_number, ETag=put_chunk_response.headers['ETag']), len(chunk)
                else:
                    logger_no_user_data.warning(
                        f'Got response with status {put_chunk_response.status_code} when uploading part {part_number}. '
                        'Retrying...'
                    )
                    logger.debug(f'Response content: {put_chunk_response.content.decode()}')

            except Exception as error:  # pylint: disable=broad-except
                logger_no_user_data.warning(f'Encountered error when uploading part {part_number}. Retrying...')
                logger.debug(f'Upload error: {error}')

            time.sleep(index * index + 2)

        logger_no_user_data.debug(f'Max retries hit, when uploading part {part_number}. Exiting...')
        raise BioLibError(f'Max retries hit, when uploading part {part_number}. Exiting...')

    def _update_progress_bar_and_parts(
        self,
        upload_chunk_response: _UploadChunkReturnType,
        parts: List[_PartMetadata],
        payload_size_in_bytes: int,
    ) -> None:
        part_metadata, chunk_byte_length = upload_chunk_response
        part_number = part_metadata['PartNumber']

        parts.append(part_metadata)
        self._bytes_uploaded += chunk_byte_length

        approx_progress_percent = min(self._bytes_uploaded / (payload_size_in_bytes + 1) * 100, 100)
        approx_rounded_progress = round(approx_progress_percent, 2)
        logger_no_user_data.debug(
            f'Uploaded part number {part_number} with size {chunk_byte_length} bytes, '
            f'the approximate progress is {approx_rounded_progress}%'
        )



import sys
import click

from biolib import biolib_errors, utils
from biolib.app import BioLibApp
from biolib.typing_utils import Tuple


@click.command(
    context_settings=dict(ignore_unknown_options=True, allow_interspersed_args=False),
    help='Run an application on BioLib',
)
@click.option('--local', is_flag=True, required=False, help='Run the application locally')
@click.option('--non-blocking', is_flag=True, required=False, help='Run the application non blocking')
@click.argument('uri', required=True)
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
def run(local: bool, non_blocking: bool, uri: str, args: Tuple[str]) -> None:
    try:
        app = BioLibApp(uri=uri)
    except biolib_errors.BioLibError as error:
        print(f'An error occurred:\n {error.message}', file=sys.stderr)
        exit(1)

    def _get_stdin():
        stdin = None
        if not sys.stdin.isatty() and not utils.IS_DEV:
            stdin = sys.stdin.read()
        return stdin

    blocking = False if non_blocking else True
    job = app.cli(
        args=list(args),
        stdin=_get_stdin(),
        files=None,
        machine=('local' if local else ''),
        blocking=blocking
    )

    if blocking:
        job.save_files('biolib_results')

        # Write stdout and stderr if it has not been streamed (Markdown is not streamed)
        if app.version.get('stdout_render_type') == 'markdown' or not sys.stdout.isatty():
            sys.stdout.buffer.write(job.get_stdout())
            sys.stderr.buffer.write(job.get_stderr())

        exit(job.get_exit_code())
    else:
        print('{"job_id": "' + str(job.id) + '"}')
